/*****************************************************************************
 * SC598_aio_Core0.c
 *****************************************************************************/
#include <sys/platform.h>
#include <sys/adi_core.h>
#include "adi_initialize.h"

#include <services/gpio/adi_gpio.h>
#include <services/pwr/adi_pwr.h>
#include <services/tmr/adi_tmr.h>
#include <services/int/adi_gic.h>

#include "ConfigSoftSwitches_EV_SC59x_SOM.h"
#include "user_usb_lib.h"

#define DELAY_COUNT             (100000u)       /*!< Delay value used to reset the ADC/DAC. */

#define CGU_DEV                 (0)             /*!< Clock Generation Unit device number. */
#define MHZ                     (1000000u)      /*!< Megahertz multiplier */
#define CLKIN                   (25u * MHZ)     /*!< 25MHz clock-in */
#define CORE_MAX                (1200u * MHZ)    /*!< 1200MHz Core Clock */
#define SYSCLK_MAX              (500u * MHZ)    /*!< 500 SysClk */
#define SCLK_MAX                (125u * MHZ)    /*!< 125 MHz SCLK */

/**
 * Mainline state machine states.
 */
typedef enum
{
    MAIN_STATE_SYSTEM_INIT,                     /*!< Initialize the system */
    MAIN_STATE_USER_INIT,                       /*!< Initialize the User code */
    MAIN_STATE_RUN,                             /*!< Runtime state */
    MAIN_STATE_ERROR                            /*!< Error State */
} Main_States;

unsigned long GpioMemory[ADI_GPIO_CALLBACK_MEM_SIZE];   /*!< ADI System Services GPIO memory*/
unsigned long gpioMaxCallbacks;                         /*!< ADI System Services GPIO max callbacks variable*/
unsigned char TimerMemory[ADI_TMR_MEMORY];
unsigned char LPTimerMemory[ADI_TMR_MEMORY];

extern void ConfigSoftSwitches(void);
static unsigned long PowerServiceInit(void);
static unsigned long GpioInit(void);
static void timer_init (void);

int main(int argc, char *argv[])
{
    Main_States main_state = MAIN_STATE_SYSTEM_INIT;
    User_Init_Return_Code rv;

    while (1)
    {
        switch (main_state)
        {
            case MAIN_STATE_SYSTEM_INIT:
                adi_initComponents(); /* auto-generated code */

                /**
                     * The default startup code does not include any functionality to allow
                     * core 0 to enable core 1 and core 2. A convenient way to enable
                     * core 1 and core 2 is to use the adi_core_enable function.
                     */
                adi_core_enable(ADI_CORE_SHARC0);
                adi_core_enable(ADI_CORE_SHARC1);

                if (PowerServiceInit() != 0)
                {
                    main_state = MAIN_STATE_ERROR;
                }
                else
                {
                    timer_init();
                    /* Software Switch Configuration for the EZ-BOARD */
                    ConfigSoftSwitches_EV_SOM();

                    if (GpioInit() != 0)
                    {
                        main_state = MAIN_STATE_ERROR;
                    }
                    else
                    {
                        main_state = MAIN_STATE_USER_INIT;
                    }
                }
            break;
            case MAIN_STATE_USER_INIT:
                rv = user_usb_lib_init();

                if (rv == USER_INIT_SUCCESS)
                {
                    main_state = MAIN_STATE_RUN;
                }
                else if (rv == USER_INIT_FAILED)
                {
                    main_state = MAIN_STATE_ERROR;
                }
            break;

            case MAIN_STATE_RUN:
                user_usb_lib_main();
            break;

            case MAIN_STATE_ERROR:

            break;
        }
    }

    return 0;
}

/**
  * Initializes ADI power system service.
  * @return 0 = Success. 1 = Error
  */
static unsigned long PowerServiceInit(void)
{
    unsigned long Result = 0u;

    if((uint32_t)adi_pwr_Init(CGU_DEV, CLKIN) != 0u)
    {
        /* return error */
        return 1u;
    }
    if((uint32_t)adi_pwr_SetPowerMode(CGU_DEV, ADI_PWR_MODE_FULL_ON) != 0u)
    {
        /* return error */
        return 1u;
    }
    return Result;
}


  /**
 * @brief Initializes GPIO service.
 *
 * @details
 * Initializes the ADI GPIO system service, and resets the ADAU1962A and ADAU1979 Codecs.
 * @return 0 = Success. 1 = Error
 */
static unsigned long GpioInit(void)
{
    unsigned long Result = 0u;

    return Result;
}

 /**
* @brief Initializes Timer service.
*
* @details
* Initializes the ADI timer system service, and configures the timer to
* generate a interrupt every 125us.
*/
static void timer_init (void)
{
    ADI_TMR_HANDLE   hTimer;
    ADI_TMR_HANDLE   hTimer2;
    ADI_TMR_RESULT   eTmrResult;

    /* Configure a 1ms Timer using the ADI system services */

    /* Open the timer */
    if( (eTmrResult = adi_tmr_Open (0,
                        TimerMemory,
                        ADI_TMR_MEMORY,
                        user_usb_lib_timer_handler,
                        0,//NULL,
                        &hTimer)) != ADI_TMR_SUCCESS)
     {
         /* Failed to open the timer handle the error here */
     }

     /*
     * Use the GP timer's API's to configure and enable the timer
     *
     */

    /* Set the mode to PWM OUT */
    eTmrResult = adi_tmr_SetMode(hTimer, ADI_TMR_MODE_CONTINUOUS_PWMOUT);

    /* Set the IRQ mode to get interrupt after timer counts to Delay + Width */
    eTmrResult = adi_tmr_SetIRQMode(hTimer, ADI_TMR_IRQMODE_WIDTH_DELAY);

    /* Set the Period */
    eTmrResult = adi_tmr_SetPeriod(hTimer, (SCLK_MAX / MHZ * 125));//15625);

    /* Set the timer width */
    eTmrResult = adi_tmr_SetWidth(hTimer, (SCLK_MAX / MHZ * 125)/2);

    /* Set the timer delay */
    eTmrResult = adi_tmr_SetDelay(hTimer, 100);

    /* Enable the timer */
    eTmrResult = adi_tmr_Enable(hTimer, true);


    /* Configure a 1ms Timer using the ADI system services */

    /* Open the timer */
    if( (eTmrResult = adi_tmr_Open (1,
                        LPTimerMemory,
                        ADI_TMR_MEMORY,
                        user_usb_lib_low_priority_timer_handler,
                        0,//NULL,
                        &hTimer2)) != ADI_TMR_SUCCESS)
     {
         /* Failed to open the timer handle the error here */
     }

     /*
     * Use the GP timer's API's to configure and enable the timer
     *
     */

    /* Set the mode to PWM OUT */
    eTmrResult = adi_tmr_SetMode(hTimer2, ADI_TMR_MODE_CONTINUOUS_PWMOUT);

    /* Set the IRQ mode to get interrupt after timer counts to Delay+Width */
    eTmrResult = adi_tmr_SetIRQMode(hTimer2, ADI_TMR_IRQMODE_PERIOD);

    /* Set the Period */
    eTmrResult = adi_tmr_SetPeriod(hTimer2, (112000));

    /* Set the timer width */
    eTmrResult = adi_tmr_SetWidth(hTimer2, 112000/2);

    /* Set the timer delay */
    eTmrResult = adi_tmr_SetDelay(hTimer2, 100);

    /* Enable the timer */
    eTmrResult = adi_tmr_Enable(hTimer2, true);
}

