/*==============================================================================
    FILE:           user_usb_lib.c

    DESCRIPTION:

    Copyright (c) 2023 Closed Loop Design, LLC

    This software is supplied "AS IS" without any warranties, express, implied
    or statutory, including but not limited to the implied warranties of fitness
    for purpose, satisfactory quality and non-infringement. Closed Loop Design LLC
    extends you a royalty-free right to use, reproduce and distribute this source
    file as well as executable files created using this source file for use with
    Analog Devices SC5xx family processors only. Nothing else gives you
    the right to use this source file.

==============================================================================*/
/*!
 * @file      user_usb_lib.c
 * @brief
 *
 * @details
 *
 *
 *            Copyright (c) 2023 Closed Loop Design, LLC
 *
 *            This software is supplied "AS IS" without any warranties, express, implied
 *            or statutory, including but not limited to the implied warranties of fitness
 *            for purpose, satisfactory quality and non-infringement. Closed Loop Design LLC
 *            extends you a royalty-free right to use, reproduce and distribute this source
 *            file as well as executable files created using this source file for use with
 *            Analog Devices SC5xx family processors only. Nothing else gives you
 *            the right to use this source file.
 *
 */

#include "user_usb_lib.h"

#include "cld_sc598_cdc_lib.h"
#include <services/gpio/adi_gpio.h>
#include <services/int/adi_int.h>
#include <services/tmr/adi_tmr.h>
#include <string.h>
#include <stdio.h>

#define SWAP_ENDIAN_4_BYTES(x)   (  (x << 24)                |  \
                                    ((x <<  8) & 0x00ff0000)  | \
                                    ((x >>  8) & 0x0000ff00)  | \
                                     (x >> 24)                  \
                                  )

#define USER_PRODUCT_ID_CDC     0x0007

/* Function prototypes */
static void user_cdc_usb_in_transfer_complete (void);
static void user_cdc_usb_in_transfer_aborted (void);
static void user_cdc_usb_out_transfer_aborted (void);
static CLD_USB_Data_Received_Return_Type user_cdc_serial_data_rx_transfer_complete (unsigned int num_bytes);
static void user_cdc_serial_data_receive (void);
static CLD_USB_Transfer_Request_Return_Type user_cdc_cmd_send_encapsulated_cmd (CLD_USB_Transfer_Params * p_transfer_data);
static CLD_USB_Transfer_Request_Return_Type user_cdc_cmd_get_encapsulated_resp (CLD_USB_Transfer_Params * p_transfer_data);
static CLD_USB_Data_Received_Return_Type user_cdc_cmd_set_line_coding (CLD_CDC_Line_Coding * p_line_coding);
static CLD_RV user_cdc_cmd_get_line_coding (CLD_CDC_Line_Coding * p_line_coding);
static CLD_USB_Data_Received_Return_Type user_cdc_cmd_set_control_line_state (CLD_CDC_Control_Line_State * p_control_line_state);
static CLD_USB_Data_Received_Return_Type user_cdc_cmd_send_break (unsigned short duration);
static CLD_USB_Data_Transmit_Return_Type user_cdc_tx_serial_data (unsigned long length, unsigned char * p_buffer);

static CLD_RV user_init_usb_phy (void);
static void user_usb_event (CLD_USB_Event event);

static void user_usb_lib_usb_isr(uint32_t Event, void *pArg);

static void user_usb_lib_status (unsigned short status_code, void * p_additional_data, unsigned short additional_data_size);

/*!< CDC Notification Interrupt IN endpoint parameters. */
static CLD_CDC_Notification_Endpoint_Params user_cdc_notification_ep_params =
{
    .endpoint_number                = 4,
    .max_packet_size_full_speed     = 64,
    .polling_interval_full_speed    = 1,
    .max_packet_size_high_speed     = 64,
    .polling_interval_high_speed    = 4,  /* 1ms */
};

/*!< CDC Serial Data Bulk OUT endpoint parameters. */
static CLD_Serial_Data_Bulk_Endpoint_Params user_cdc_serial_data_rx_ep_params =
{
    .endpoint_number                = 5,
    .max_packet_size_full_speed     = 64,
    .max_packet_size_high_speed     = 512,
};

/*!< CDC Serial Data Bulk IN endpoint parameters. */
static CLD_Serial_Data_Bulk_Endpoint_Params user_cdc_serial_data_tx_ep_params =
{
    .endpoint_number                = 5,
    .max_packet_size_full_speed     = 64,
    .max_packet_size_high_speed     = 512,
};


/*!< CLD SC59x Audio 2.0 library initialization data. */
static CLD_SC598_CDC_Lib_Init_Params user_usb_lib_init_params =
{
    .vendor_id                     = 0x064b,       /* Analog Devices Vendor ID */
    .product_id                    = USER_PRODUCT_ID_CDC,

    .usb_bus_max_power             = 0,
    .device_descriptor_bcdDevice   = 0x0100,


    .phy_hs_timeout_calibration     = 48/8, /* Round 43 so it is evenly divisable by 8. */
    .phy_fs_timeout_calibration     = 0,
    .phy_delay_req_after_ulip_chirp_cmd = CLD_TRUE,

    .fp_init_usb_phy                = user_init_usb_phy,

    .p_serial_data_rx_endpoint_params                   = &user_cdc_serial_data_rx_ep_params,
    .p_serial_data_tx_endpoint_params                   = &user_cdc_serial_data_tx_ep_params,
    .p_notification_endpoint_params                     = &user_cdc_notification_ep_params,

    .fp_cdc_cmd_send_encapsulated_cmd                   = user_cdc_cmd_send_encapsulated_cmd,
    .fp_cdc_cmd_get_encapsulated_resp                   = user_cdc_cmd_get_encapsulated_resp,

    .fp_cdc_cmd_set_line_coding                         = user_cdc_cmd_set_line_coding,
    .fp_cdc_cmd_get_line_coding                         = user_cdc_cmd_get_line_coding,

    .fp_cdc_cmd_set_control_line_state                  = user_cdc_cmd_set_control_line_state,

    .fp_cdc_cmd_send_break                              = user_cdc_cmd_send_break,

    .support_cdc_network_connection                     = 1,
    .cdc_class_bcd_version                              = 0x0120,       /* CDC Version 1.2 */
    .cdc_class_control_protocol_code                    = 0,            /* No Class Specific protocol */

    .p_usb_string_manufacturer                          = "Analog Devices Inc",
    .p_usb_string_product                               = "SC598 CDC/ACM Device",
    .p_usb_string_serial_number                         = CLD_NULL,
    .p_usb_string_configuration                         = CLD_NULL,
    .p_usb_string_communication_class_interface         = "CLD CDC Ctrl",
    .p_usb_string_data_class_interface                  = "CLD CDC Data",

    .user_string_descriptor_table_num_entries           = 0,
    .p_user_string_descriptor_table                     = NULL,

    .usb_string_language_id                             = 0x0409,       /* English (US) language ID */

    .fp_cld_usb_event_callback                          = user_usb_event,

    .fp_cld_lib_status                                  = user_usb_lib_status,
};

#pragma pack(1)

/* Current CDC line coding settings (default to 115.2k 8 bits, 1 Stop bit, no parity) */
static CLD_CDC_Line_Coding user_cdc_line_coding __attribute__ ((aligned (4), section(".l3_uncached_data")))=
{
    .data_terminal_rate = 115200,
    .num_stop_bits      = CLD_CDC_LINE_CODING_1_STOP_BITS,
    .parity             = CLD_CDC_LINE_CODING_PARITY_NONE,
    .num_data_bits      = 8,
};


#define USER_CDC_SERIAL_BUFFER_SIZE     512                 /*!< CDC Serial Data receive buffer size */
/**
 * Buffer used to store the received serial data.
 */
typedef struct
{
    unsigned char resp_hdr[12]; /* Must be a multiple of 4-bytes to maintain USB library required 4-byte alignment */
    unsigned char data[USER_CDC_SERIAL_BUFFER_SIZE];
} CDC_Serial_Data;
static CDC_Serial_Data user_cdc_serial_data_rx_buffer __attribute__ ((aligned (4), section(".l3_uncached_data"))) =
{
    .resp_hdr = "\n\rLib Echo: ",
};

/**
 * Initializes the CLD USB library.
 *
 * @retval USER_AUDIO_INIT_SUCCESS.
 * @retval USER_AUDIO_INIT_ONGOING.
 * @retval USER_AUDIO_INIT_FAILED.
 */
User_Init_Return_Code user_usb_lib_init (void)
{
    static unsigned char state = 0;
    User_Init_Return_Code rv = USER_INIT_ONGOING;
    CLD_RV cld_rv;

    switch (state)
    {
        case 0:
            /* Configure the GPIO pins connected to the SC589 EZ-Board general purpose LEDs (LED10, LED11, LED12) */
            adi_gpio_SetDirection(ADI_GPIO_PORT_C, ADI_GPIO_PIN_3, ADI_GPIO_DIRECTION_OUTPUT);
            adi_gpio_Clear(ADI_GPIO_PORT_C, ADI_GPIO_PIN_3);
            adi_gpio_SetDirection(ADI_GPIO_PORT_C, ADI_GPIO_PIN_2, ADI_GPIO_DIRECTION_OUTPUT);
            adi_gpio_Clear(ADI_GPIO_PORT_C, ADI_GPIO_PIN_2);
            adi_gpio_SetDirection(ADI_GPIO_PORT_C, ADI_GPIO_PIN_1, ADI_GPIO_DIRECTION_OUTPUT);
            adi_gpio_Clear(ADI_GPIO_PORT_C, ADI_GPIO_PIN_1);

            state++;
        break;
        case 1:
            /* Initialize the CLD Audio v2.0 with CDC Library */
            cld_rv = cld_sc598_cdc_lib_init(&user_usb_lib_init_params);

            if (cld_rv == CLD_SUCCESS)
            {
                adi_int_InstallHandler(INTR_USBC0_INT, user_usb_lib_usb_isr, NULL, 1);

                /* Connect to the USB Host */
                cld_lib_usb_connect();


                rv = USER_INIT_SUCCESS;
            }
            else if (cld_rv == CLD_FAIL)
            {
                rv = USER_INIT_FAILED;
            }
            else
            {
                rv = USER_INIT_ONGOING;
            }
        break;
    }
    return rv;
}


/**
 * Example User Mainline function.
 *
 */
void user_usb_lib_main (void)
{
    static CLD_Time main_time = 0;

    if (cld_time_passed_ms(main_time) >= 250u)
    {
        main_time = cld_time_get();
        adi_gpio_Toggle(ADI_GPIO_PORT_C, ADI_GPIO_PIN_3);
    }
    cld_sc598_cdc_lib_main();
}

/**
 * This function is called by the CLD library when Bulk IN serial data has been
 * transmitted to the Host.
 */
static void user_cdc_usb_in_transfer_complete (void)
{
}
/**
 * This function is called by the CLD library when Bulk IN serial data has been
 * aborted.
 */
static void user_cdc_usb_in_transfer_aborted (void)
{
}

/**
 * Function called when Serial data has been received.
 *
 * @retval CLD_USB_DATA_GOOD
 * @retval CLD_USB_DATA_BAD_STALL
 */
static CLD_USB_Data_Received_Return_Type user_cdc_serial_data_rx_transfer_complete (unsigned int num_bytes)
{
    /* Transmit the received data using the SC59x UART. */
    user_cdc_tx_serial_data(num_bytes+sizeof(user_cdc_serial_data_rx_buffer.resp_hdr), (unsigned char*)&user_cdc_serial_data_rx_buffer);
    user_cdc_serial_data_receive();

    return CLD_USB_DATA_GOOD;
}

/**
 * This function is called by the CLD library when Bulk OUT
 * serial data has been aborted.
 */
static void user_cdc_usb_out_transfer_aborted (void)
{
    user_cdc_serial_data_receive();
}

/**
 * This function primes the Bulk OUT endpoint so it will
 * receive data.
 */
static void user_cdc_serial_data_receive (void)
{
    static CLD_USB_Transfer_Params transfer_params;
    CLD_USB_Data_Receive_Return_Type rv;

    transfer_params.num_bytes = USER_CDC_SERIAL_BUFFER_SIZE;
    transfer_params.p_data_buffer = user_cdc_serial_data_rx_buffer.data;
    transfer_params.transfer_timeout_ms = 0;
    transfer_params.fp_transfer_aborted_callback = user_cdc_usb_out_transfer_aborted;
    transfer_params.callback.fp_usb_out_transfer_complete = user_cdc_serial_data_rx_transfer_complete;

    rv = cld_cdc_lib_receive_serial_data(&transfer_params);
    if (rv == CLD_USB_RECEIVE_FAILED_MISALIGNED)
    {
        /* CLD library requires transfer_params.p_data_buffer to be 32-bit aligned */
    }
    else if (rv == CLD_USB_RECEIVE_FAILED)
    {
        /* Bulk OUT endpoint busy */
    }
}

/**
 * This function is called by the CLD library when a CDC Send Encapsulated Command
 * request is received. This function sets the p_transfer_data parameters to
 * select where the received data should be stored, and what function should
 * be called when the transfer is complete.
 *
 * @param p_transfer_data                - Pointer to the Control OUT transfer data.
 * @param p_transfer_data->num_bytes     - Number of Control OUT bytes.
 * @param p_transfer_data->p_data_buffer - Set to the address where to
 *                                         store the Control OUT data.
 * @param p_transfer_data->callback.usb_in_transfer_complete -
 *                                         Function called when the
 *                                         Control OUT bytes have been
 *                                         received.
 * @param p_transfer_data->transfer_aborted_callback -
 *                                         Optional function that is
 *                                         called if the transfer is
 *                                         aborted.
 * @retval CLD_USB_TRANSFER_ACCEPT - Store the data using the p_transfer_data
 *                                  parameters.
 * @retval CLD_USB_TRANSFER_PAUSE - The device isn't ready to process this
 *                                  out packet so pause the transfer
 *                                  until the cld_sc57x_audio_2_0_w_cdc_lib_resume_paused_control_transfer
 *                                  function is called.
 * @retval CLD_USB_TRANSFER_DISCARD - Discard this packet.
 * @retval CLD_USB_TRANSFER_STALL - Stall the OUT endpoint.
 */
static CLD_USB_Transfer_Request_Return_Type user_cdc_cmd_send_encapsulated_cmd (CLD_USB_Transfer_Params * p_transfer_data)
{
    /* Not currently supported by this example. */
    return CLD_USB_TRANSFER_STALL;
}

/**
 * This function is called by the CLD library when a CDC Get Encapsulated Command
 * request is received. This function sets the p_transfer_data parameters to
 * select where the transmit data should be sourced, and what function should
 * be called when the transfer is complete.
 *
 * @param p_transfer_data                - Pointer to the Control IN transfer data.
 * @param p_transfer_data->num_bytes     - Number of Control IN bytes.
 * @param p_transfer_data->p_data_buffer - Set to the address where to
 *                                         source the Control IN data.
 * @param p_transfer_data->callback.usb_in_transfer_complete -
 *                                         Function called when the
 *                                         Control IN bytes have been
 *                                         sent.
 * @param p_transfer_data->transfer_aborted_callback -
 *                                         Optional function that is
 *                                         called if the transfer is
 *                                         aborted.
 * @retval CLD_USB_TRANSFER_ACCEPT - Store the data using the p_transfer_data
 *                                   parameters.
 * @retval CLD_USB_TRANSFER_PAUSE - The device isn't ready to process this
 *                                  out packet so pause the transfer
 *                                  until the cld_sc57x_audio_2_0_w_cdc_lib_resume_paused_control_transfer
 *                                  function is called.
 * @retval CLD_USB_TRANSFER_DISCARD - Discard this packet.
 * @retval CLD_USB_TRANSFER_STALL - Stall the OUT endpoint.
 */
static CLD_USB_Transfer_Request_Return_Type user_cdc_cmd_get_encapsulated_resp (CLD_USB_Transfer_Params * p_transfer_data)
{
    /* Not currently supported by this example. */
    return CLD_USB_TRANSFER_STALL;
}

/**
 * This function is called by the CLD library when a CDC Set Line Coding
 * request is received. This function saves the requested line coding.
 *
 * @retval CLD_USB_DATA_GOOD
 * @retval CLD_USB_DATA_BAD_STALL
 */
static CLD_USB_Data_Received_Return_Type user_cdc_cmd_set_line_coding (CLD_CDC_Line_Coding * p_line_coding)
{
    user_cdc_line_coding = *p_line_coding;
    return CLD_USB_DATA_GOOD;
}

/**
 * This function is called by the CLD library when a CDC Get Line Coding
 * request is received. This function loads the current line coding into the
 * p_line_coding structure passed to the function.
 *
 * @retval CLD_SUCCESS - There was no errors copying the current line coding.
 * @retval CLD_FAIL    - There was a problem copying the current line coding.
 */
static CLD_RV user_cdc_cmd_get_line_coding (CLD_CDC_Line_Coding * p_line_coding)
{
    *p_line_coding = user_cdc_line_coding;

    return CLD_SUCCESS;
}

/**
 * This function is called by the CLD library when a CDC Set Control Line State
 * request is received.
 *
 * @retval CLD_USB_DATA_GOOD - Set control line state parameters accepted.
 * @retval CLD_USB_DATA_BAD_STALL - An error occurred while applying the requested
 *                                  set control line state parameters.
 */
static CLD_USB_Data_Received_Return_Type user_cdc_cmd_set_control_line_state (CLD_CDC_Control_Line_State * p_control_line_state)
{
    return CLD_USB_DATA_GOOD;
}

/**
 * This function is called by the CLD library when a CDC Send Break
 * request is received.
 *
 * @retval CLD_USB_DATA_GOOD
 * @retval CLD_USB_DATA_BAD_STALL
 */
static CLD_USB_Data_Received_Return_Type user_cdc_cmd_send_break (unsigned short duration)
{
    return CLD_USB_DATA_GOOD;
}

/**
 * Transmits the specified data to the CDC Host using the serial data Bulk IN endpoint.
 *
 * @param length   - number of bytes to transmit.
 * @param p_buffer - pointer to the data buffer to transmit
 */
static CLD_USB_Data_Transmit_Return_Type user_cdc_tx_serial_data (unsigned long length, unsigned char * p_buffer)
{
    static CLD_USB_Transfer_Params transfer_params;
    transfer_params.num_bytes = length;
    transfer_params.p_data_buffer = p_buffer;
    transfer_params.callback.fp_usb_in_transfer_complete = user_cdc_usb_in_transfer_complete;
    transfer_params.fp_transfer_aborted_callback = user_cdc_usb_in_transfer_aborted;
    transfer_params.transfer_timeout_ms = 1000;

    return cld_cdc_lib_transmit_serial_data (&transfer_params);
}

/**
 * Function called when a USB event occurs on the USB Port with USB Audio v2.0 and CDC.
 *
 * @param event - identifies which USB event has occurred.
 */
static void user_usb_event (CLD_USB_Event event)
{
    switch (event)
    {
        case CLD_USB_CABLE_CONNECTED:
            adi_gpio_Set(ADI_GPIO_PORT_C, ADI_GPIO_PIN_1);
        break;
        case CLD_USB_CABLE_DISCONNECTED:
            adi_gpio_Clear(ADI_GPIO_PORT_C, ADI_GPIO_PIN_1);
        break;
        case CLD_USB_ENUMERATED_CONFIGURED_HS:
        case CLD_USB_ENUMERATED_CONFIGURED_FS:
            user_cdc_serial_data_receive();
        break;
        case CLD_USB_UN_CONFIGURED:
        break;
        case CLD_USB_BUS_RESET:
        break;
    }
}

/**
 * User defined USB 0 ISR.
 *
 * @param Event, pArg.
 */
static void user_usb_lib_usb_isr(uint32_t Event, void *pArg)
{
    cld_usb0_isr_callback();
}

/**
 * User defined Timer ISR.
 *
 * @param pCBParam, Event, pArg.
 */
void user_usb_lib_timer_handler(void *pCBParam, uint32_t Event, void *pArg)
{
    switch(Event)
    {
        case ADI_TMR_EVENT_DATA_INT:
            /* Call the CLD library 125 microsecond function */
            cld_time_125us_tick();
            break;

        default:
            break;
    }

    return;
}
/**
 * User defined Timer ISR - currently set to run every millisecond.
 *
 * @param pCBParam, Event, pArg.
 */
void user_usb_lib_low_priority_timer_handler(void *pCBParam, uint32_t Event, void *pArg)
{
    switch(Event)
    {
        case ADI_TMR_EVENT_DATA_INT:

            break;

        default:
            break;
    }

    return;
}

/**
 * Function called when the CLD library reports a status.
 *
 * @param status_code, p_additional_data, additional_data_size.
 */
static void user_usb_lib_status (unsigned short status_code, void * p_additional_data, unsigned short additional_data_size)
{
    /* Process the CLD library status */
    //char * p_str = cld_lib_status_decode(status_code, p_additional_data, additional_data_size);
}

/**
 * Function called when the system services uart receives a byte.
 *
 * @param byte.
 */
void user_usb_lib_rx_term_byte (unsigned char byte)
{
    switch (byte)
    {

    }
}

/**.
 *  Function used to initialize and reset the USB Phy.
 * @return CLD_ONGOING results in this function getting additional runtime.
 * @return CLD_SUCCESS USB Phy initialized successfully.
 * @return CLD_FAIL Phy initialization failed, USB library initialization failure
 */
static CLD_RV user_init_usb_phy (void)
{
    static unsigned char phy_init_state = 0;
    static CLD_Time phy_init_time;
    CLD_RV rv = CLD_ONGOING;

    switch (phy_init_state)
    {
        case 0:
            /* Configure the PHY reset PIN */
            adi_gpio_SetDirection(ADI_GPIO_PORT_G, ADI_GPIO_PIN_11, ADI_GPIO_DIRECTION_OUTPUT);
            /* Assert the reset PIN */
            adi_gpio_Clear(ADI_GPIO_PORT_G, ADI_GPIO_PIN_11);

            phy_init_time = cld_time_get();
            phy_init_state++;
        break;
        case 1:
            if (cld_time_passed_ms(phy_init_time) > 10)
            {
                /* Unassert reset */
                adi_gpio_Set(ADI_GPIO_PORT_G, ADI_GPIO_PIN_11);
                phy_init_time = cld_time_get();
                phy_init_state++;
            }
        break;
        case 2:
            /* After releasing reset the USB3340 has a Tstart time of 1.2ms Max */
            if (cld_time_passed_ms(phy_init_time) > 10)
            {
                rv = CLD_SUCCESS;
            }
        break;
    }
    return rv;
}

